import time
import asyncio
import threading
import logging
from abc import abstractmethod, ABCMeta
from . import config
from .core import REPEAT_QUEUE


class LogParser(metaclass=ABCMeta):

    @abstractmethod
    def parse(self, line):
        """Implement the method in subclass and return an instance of RequestInfo

        :param line: one line in log file
        :type line: str
        :return: an instance of RequestInfo
        """
        pass


class ParserThread(threading.Thread):

    def __init__(self, log_file, log_parser, file_encoding='utf-8'):
        super(ParserThread, self).__init__()
        if not issubclass(log_parser, LogParser):
            raise TypeError
        self.log_file = log_file
        self.log_parser = log_parser()
        self.logger = logging.getLogger(__name__)
        self.file_encoding = file_encoding
        self.loop = asyncio.get_event_loop()

    def run(self):
        last_gather_ts = None
        diff_ts = None  # 回放时间与原记录时间差
        matched_records_in_cycle = 0

        with open(self.log_file, "r", encoding=self.file_encoding) as f:
            for line in f.readlines():
                request_info = self.log_parser.parse(line)
                if not request_info.is_matched:  # 当该行并非请求日志时,不处理
                    continue

                if config.GATHER_INTERVAL == 0:  # 不控制采集间隔的情况下, 不会控制回放的节奏
                    self.logger.debug(request_info)
                    self.loop.call_soon_threadsafe(REPEAT_QUEUE.put_nowait, request_info)
                else:
                    if request_info.timestamp is None:
                        raise ValueError("if set GATHER_INTERVAL not eq 0, must specify the `timestamp` "
                                         "in the instance of `RequestInfo`")
                    if last_gather_ts is None:  # 取到第一条匹配的请求日志, 回放开始
                        last_gather_ts = request_info.timestamp
                        diff_ts = int(time.time() * 1000) - last_gather_ts
                        matched_records_in_cycle += 1
                        self.logger.debug(request_info)
                        self.loop.call_soon_threadsafe(REPEAT_QUEUE.put_nowait, request_info)
                        continue

                    if request_info.timestamp - last_gather_ts <= config.GATHER_INTERVAL * 1000:
                        self.logger.debug(request_info)
                        matched_records_in_cycle += 1
                        self.loop.call_soon_threadsafe(REPEAT_QUEUE.put_nowait, request_info)
                    else:
                        while 1:
                            self.logger.info("gathered {} records in {} seconds".format(
                                matched_records_in_cycle, config.GATHER_INTERVAL))
                            time.sleep(config.GATHER_INTERVAL)
                            last_gather_ts = int(time.time()*1000) - diff_ts
                            matched_records_in_cycle = 0

                            if request_info.timestamp - last_gather_ts < config.GATHER_INTERVAL * 1000:
                                matched_records_in_cycle += 1
                                self.logger.debug(request_info)
                                self.loop.call_soon_threadsafe(REPEAT_QUEUE.put_nowait, request_info)
                                break

        self.logger.info("read complete")



import asyncio
import timeit
from asyncio import Queue
import random
import logging
import aiohttp
from . import events
from . import config
from .model import RequestInfo


EVENT_LOOP = asyncio.get_event_loop()
REPLAY_QUEUE = Queue(loop=EVENT_LOOP)  # 回放队列,生产者为中继器,消费者将会生成的异步请求
REPEAT_QUEUE = Queue(loop=EVENT_LOOP)  # 中继队列,消费者为中继器,用于更改请求量
CLIENT = aiohttp.ClientSession(loop=EVENT_LOOP)
LOGGER = logging.getLogger(__name__)


async def repeater(rate):
    """中继器

    :param rate:
    :return:
    """
    while 1:
        request_info = await REPEAT_QUEUE.get()
        replay_rate = rate

        if request_info == config.FINISHED_SIGNAL:
            for _ in range(int(round((config.PLAYER_NUMBER/config.REPEATER_NUMBER)+0.5))):
                REPLAY_QUEUE.put_nowait(request_info)  # notify the player to finish
            LOGGER.info("repeat finished")
            break

        if not isinstance(request_info, RequestInfo):
            LOGGER.error("incorrect message in repeat queue: {}".format(request_info))
            continue

        while replay_rate > 0:
            if replay_rate >= 1:
                parameters = request_info.to_request_parameters()
                events.repeat.fire(parameters=parameters)  # do not unpack parameters
                REPLAY_QUEUE.put_nowait(parameters)
            else:
                r = random.random()
                if r <= rate:
                    parameters = request_info.to_request_parameters()
                    events.repeat.fire(parameters=parameters)  # do not unpack parameters
                    REPLAY_QUEUE.put_nowait(parameters)
            replay_rate -= 1


async def request(method, url, **kwargs):
    """http请求函数

    :param method:
    :param url:
    :param kwargs:
    :return:
    """
    future = asyncio.Future()
    if config.CALLBACK:
        future.add_done_callback(config.CALLBACK)

    start = timeit.default_timer()
    async with CLIENT.request(method, url, **kwargs) as response:
        content = await response.text()
        result = dict(
            request=kwargs,
            response=dict(
                status_code=response.status,
                content=content,
                latency=int((timeit.default_timer()-start)*1000),
                method=response.method,
                url=response.url,
                headers=response.headers,
                cookies=response.cookies
            ))
        if config.CALLBACK:
            future.set_result(result)


async def player():
    """请求发送器

    :param q:
    :return:
    """
    while 1:
        parameters = await REPLAY_QUEUE.get()
        if parameters == config.FINISHED_SIGNAL:
            LOGGER.info("player finished")
            break

        events.replay.fire(parameters=parameters)

        method = parameters.pop('method', 'get')
        url = parameters.pop('url', None)
        params = parameters.pop('params', None)
        data = parameters.pop('data', None)
        headers = parameters.pop('headers', None)
        try:
            await request(method, url, params=params, data=data, headers=headers, **parameters)
        except:
            pass



import os
import asyncio
from concurrent.futures import ThreadPoolExecutor
from . import config
from .core import player, repeater
from .parser import ParserThread
from .monitor import MonitorThread

__all__ = ["main"]


def main(log_file, log_parser, rate=1, file_encoding="utf-8", callback=None):
    """log replay主函数

    :param log_file: log file path
    :param rate: 默认原速回; 当>1时, 比如2,从日志中读取到的一条http请求,将会被并发请求两次; 当0<rate<1时,如0.5,该http请求,有50%的概率不回放
    :param log_parser: 日志解析器
    :param file_encoding: 日志文件编码
    :param callback:
    :return:
    """

    if not os.path.isfile(log_file):
        raise IOError("cannot find log file: {}".format(log_file))

    if callback is not None:
        config.CALLBACK = callback

    [asyncio.ensure_future(repeater(rate)) for _ in range(config.REPEATER_NUMBER)]
    [asyncio.ensure_future(player()) for _ in range(config.PLAYER_NUMBER)]

    event_loop = asyncio.get_event_loop()
    event_loop.set_default_executor(ThreadPoolExecutor(config.THREAD_POOL_NUMBER))

    pt = ParserThread(log_file, log_parser, file_encoding=file_encoding)
    pt.start()
    MonitorThread(pt).start()

    event_loop.run_forever()
    event_loop.close()



import logging
import asyncio
import threading
import time
from . import config
from .core import REPEAT_QUEUE, CLIENT, EVENT_LOOP


class MonitorThread(threading.Thread):

    def __init__(self, monitored):
        super(MonitorThread, self).__init__()
        self.monitored = monitored
        self.logger = logging

    def run(self):
        while self.monitored.isAlive():
            time.sleep(.5)
        self.logger.info("the monitored thread is dead")

        # send finished signal to all repeater
        self.logger.info('send finish signal to REPEAT_QUEUE')
        for _ in range(config.REPEATER_NUMBER):
            EVENT_LOOP.call_soon_threadsafe(REPEAT_QUEUE.put_nowait, config.FINISHED_SIGNAL)

        self.logger.info("waiting for all tasks to complete")
        while asyncio.Task.all_tasks(loop=EVENT_LOOP):
            time.sleep(.5)

        # close loop
        self.logger.info('stop event loop safety')
        EVENT_LOOP.call_soon_threadsafe(CLIENT.close)
        EVENT_LOOP.call_soon_threadsafe(EVENT_LOOP.stop)



class RequestInfo:
    url = None
    method = "get"
    headers = None
    data = None
    params = None
    cookies = None
    timestamp = None
    is_matched = False

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_request_parameters(self):
        return {k: v for k, v in self.__dict__.items() if k not in ("timestamp", "is_matched")}


__all__ = ["response", "repeat", "replay"]


class EventHook:

    def __init__(self):
        self._handlers = []

    def __iadd__(self, other):
        self._handlers.append(other)
        return self

    def __isub__(self, other):
        self._handlers.remove(other)
        return self

    def fire(self, **kwargs):
        for handler in self._handlers:
            handler(**kwargs)


repeat = EventHook()  # arguments: parameters
replay = EventHook()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='logreplay',
    version='0.2.0',
    packages=['logreplay'],
    url='https://github.com/jacexh/log-replay',
    license='MIT',
    author='jace',
    author_email='jace@xuh.me',
    description='a log-based tool to replay http requests',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
    ],
    install_requires=[
        "aiohttp",
    ]
)
